

def get_coord(string):
    lines = string.split('\n')
    line = len(lines)
    column = len(lines[-1]) + 1
    return line, column
assert(get_coord('aa') == (1, 3))


def get_all_tokens(nested_tokens):
    tokens = []
    for token in nested_tokens:
        if isinstance(token, BaseTypeContainer):
            tokens += get_all_tokens(token.tokens)
        else:
            tokens.append(token)
    return tokens


def get_diff(string):
    lines = string.split('\n')
    line = len(lines) - 1
    column = len(lines[-1])
    return line, column
assert(get_diff('') == (0, 0))
assert(get_diff('aa') == (0, 2))
assert(get_diff('aa\n') == (1, 0))
assert(get_diff('aa\na') == (1, 1))


class BaseType:
    """
    This class is used to count the string-coordinate (line, column) of any element in a statement.
    This is used for identifying, in a script, the line and column of an error.
    It also defines the __eq__
    """
    def __init__(self):
        self._position = None

    @property
    def _key(self):
        # idiom described in https://stackoverflow.com/a/2909119/931303
        return tuple(x for x in sorted(self.__dict__.items()) if x[0] != '_position')

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._key == other._key

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._key)

    def set_position(self, position):
        assert (isinstance(position, tuple))
        assert (len(position) == 2)
        self._position = position

    @property
    def undefined_position(self):
        return self._position is None

    @property
    def position(self):
        if self.undefined_position:
            raise Exception(self, type(self))
        return self._position

    @position.setter
    def position(self, position):
        self.set_position(position)


class ParserType(BaseType):
    # base type ignored by the interpreter
    pass


class BaseTypeContainer(BaseType):
    """
    This is the base class for containers (e.g. statements, code).

    Relevant of this class:
        * `base_tokens` to get tokens that have functionality.
        * `string_up_to`: the string representation of this class up to an index.
    """
    def __init__(self, tokens):
        super().__init__()
        for i, s in enumerate(tokens):
            assert(isinstance(s, BaseType))
        self._tokens = tokens

    @staticmethod
    def is_base_token(token):
        raise NotImplementedError

    def _as_str(self, func=str):
        raise NotImplementedError

    def set_position(self, position):
        self._position = position
        for token in self._tokens:
            token.set_position(position)

            token_delta = get_diff(str(token))

            if token_delta[0] == 0:
                initial_column = position[1]
            else:
                initial_column = 1

            position = (
                position[0] + token_delta[0],
                initial_column + token_delta[1]
            )

    @BaseType.position.setter
    def position(self, position):
        super().set_position(position)

    @property
    def tokens(self):
        return self._tokens

    def get_all_tokens(self):
        return get_all_tokens(self.tokens)

    @property
    def base_tokens(self):
        return [token for token in self._tokens if self.is_base_token(token)]

    def __str__(self):
        return self._as_str()



from sqf.base_type import ParserType
from sqf.keywords import BINARY_OPERATORS, UNARY_OPERATORS, OP_COMPARISON, PREPROCESSORS_UNARY


class EndToken:
    pass


def _normalize(item):
    return str(item).lower()


def nud(token, parser):
    n_token = _normalize(token)
    if n_token in UNARY_OPERATORS:
        return parser.container([token, parser.expression(100)])
    elif str(token) in PREPROCESSORS_UNARY:
        return parser.container([token, parser.expression(100)])
    elif str(token) == '#define':
        arg = parser.expression(100)
        args = parser.expression(100)
        func = parser.expression(100)
        return parser.container([token, arg, args, func])
    elif str(token).isupper():
        # heuristic to catch global defines with arguments
        if str(parser.next)[0] == '(':
            return parser.container([token, parser.expression(100)])
    return token


def get_lbp(token):
    n_token = _normalize(token)

    if token == EndToken:
        return 0
    elif n_token == '=':
        return 0.8
    elif n_token == 'private':
        return 0.9
    elif n_token in ('||', 'or'):
        return 1
    elif n_token in {'&&', 'and'}:
        return 2
    elif n_token in set(x.value for x in OP_COMPARISON):
        return 3
    elif n_token in {'*', '/', '%', 'mod', 'atan2'}:
        return 7
    elif n_token in {'+', 'max', 'min', '-'}:
        return 6
    elif n_token == 'else':
        return 5
    elif n_token in BINARY_OPERATORS:
        return 4
    elif n_token in UNARY_OPERATORS:
        return 9
    else:
        return 0.1


class Parser:

    def __init__(self, container):
        self.next = None
        self.container = container
        self.tokens = []
        self.cumulator = []
        self.iterator = self._iterator()

    def _iterator(self):
        for token in self.tokens:
            if isinstance(token, ParserType):
                self.cumulator.append(token)
            else:
                yield token
        yield EndToken

    def expression(self, rbp=0):
        current = self.next

        cum_prefix = self.cumulator
        self.cumulator = []

        try:
            self.next = next(self.iterator)
        except StopIteration:
            if len(cum_prefix + self.cumulator) == 1:
                return (cum_prefix + self.cumulator)[0]
            return self.container(cum_prefix + self.cumulator)

        left = nud(current, self)
        if cum_prefix + self.cumulator:
            left = self.container(cum_prefix + [left] + self.cumulator)
            self.cumulator = []

        while rbp < get_lbp(self.next):
            current = self.next
            self.next = next(self.iterator)
            if self.next is EndToken:
                return self.container([left, current])
            right = self.expression(get_lbp(current))
            left = self.container([left, current, right])

        return left

    def parse(self, tokens):
        if len(tokens) == 1:
            return tokens[0]
        self.tokens = tokens
        self.iterator = self._iterator()
        self.next = next(self.iterator)
        return self.expression()


def parse_exp(tokens, container=list):
    return Parser(container).parse(tokens)



from sqf.types import Code, String, Number, Array, Type, Variable, Boolean, Namespace, _Statement, Nothing, Statement


class InterpreterType(Type):
    # type that is used by the interpreter (e.g. While type)
    pass


class _InterpreterType(InterpreterType):
    def __init__(self, token):
        assert (isinstance(token, Type))
        super().__init__()
        self.token = token

    @property
    def is_undefined(self):
        return self.token.is_undefined


class PrivateType(_InterpreterType):
    """
    A type to store the result of "private _x" as in "private _x = 2"
    """
    def __init__(self, variable):
        assert(isinstance(variable, Variable))
        super().__init__(variable)

    @property
    def variable(self):
        return self.token


class WhileType(_InterpreterType):
    def __init__(self, condition):
        assert(isinstance(condition, Code))
        super().__init__(condition)

    @property
    def condition(self):
        return self.token


class ForType(_InterpreterType):
    def __init__(self, variable=None, from_=None, to=None, step=None):
        if step is None:
            step = Number(1)
        if variable is None:
            variable = String()
        assert (variable is None or isinstance(variable, String))
        assert (from_ is None or isinstance(from_, Type))
        assert (to is None or isinstance(to, Type))
        assert (isinstance(step, Type))
        super().__init__(variable)
        self.from_ = from_
        self.to = to
        self.step = step

    @property
    def variable(self):
        return self.token

    @property
    def is_undefined(self):
        return self.variable.is_undefined or \
               self.from_ is not None and self.from_.is_undefined or \
               self.to is not None and self.to.is_undefined

    def copy(self, other):
        self.token = other.variable
        self.from_ = other.from_
        self.to = other.to
        self.step = other.step


class ForSpecType(_InterpreterType):
    def __init__(self, array):
        assert (isinstance(array, Array))
        super().__init__(array)

    @property
    def array(self):
        return self.token


class SwitchType(_InterpreterType):
    def __init__(self, keyword, result):
        super().__init__(result)
        self.keyword = keyword

    @property
    def result(self):
        return self.token


class IfType(_InterpreterType):
    def __init__(self, condition=None):
        if condition is None:
            condition = Boolean()
        super().__init__(condition)

    @property
    def condition(self):
        return self.token


class ElseType(_InterpreterType):
    def __init__(self, then=None, else_=None):
        super().__init__(then)
        if then is None:
            condition = Boolean()
        assert (isinstance(then, Code))
        assert (isinstance(else_, Code))
        self.else_ = else_

    @property
    def then(self):
        return self.token

    @property
    def is_undefined(self):
        return self.else_.is_undefined or self.then.is_undefined


class TryType(_InterpreterType):
    def __init__(self, code):
        assert (isinstance(code, Code))
        super().__init__(code)


class WithType(_InterpreterType):
    def __init__(self, namespace):
        assert (isinstance(namespace, Namespace))
        super().__init__(namespace)

    @property
    def namespace(self):
        return self.token


class DefineStatement(_Statement, InterpreterType):
    def __init__(self, tokens, variable_name, expression=None, args=None):
        assert(isinstance(variable_name, str))
        assert(isinstance(tokens, list))
        super().__init__(tokens)
        self.variable_name = variable_name
        if expression is None:
            expression = [Nothing()]
        self.expression = expression
        if args is None:
            args = []
        self.args = args

    def __repr__(self):
        return '#d<%s>' % self._as_str(repr)


class IfDefStatement(_Statement, InterpreterType):

    def __init__(self, tokens, statement_class=Statement):
        super().__init__(tokens)
        self.statement_class = statement_class

    def __repr__(self):
        return '#i<%s>' % self._as_str(repr)


class DefineResult(_Statement, InterpreterType):
    """
    A statement whose some token was replaced by a #define.
    The statement contain the original tokens, but also what define_statement was used,
    and what was the resulting statement after replacement.

    str(self) still returns the original tokens, but `result` can be used to evaluate the statement.
    """
    def __init__(self, tokens, define_statement, result):
        super().__init__(tokens)
        self.define_statement = define_statement
        assert (isinstance(result, (Type, Statement)))
        self.result = result

    def __repr__(self):
        return '#dR|%s -> %s|' % (self.tokens, (''.join('%s' % self.result)).replace('\n', '\\n'))


class IfDefResult(_Statement, InterpreterType):
    def __init__(self, ifdef_statement, result):
        super().__init__(ifdef_statement.tokens)
        self.ifdef_statement = ifdef_statement
        assert(isinstance(result, list))
        self.result = result  # a list of statements

    def __repr__(self):
        return '#iR|%s -> %s|' % (self.tokens, (''.join('%s' % self.result)).replace('\n', '\\n'))



import sys
import os
import argparse

from sqf.parser import parse
import sqf.analyzer
from sqf.exceptions import SQFParserError, SQFWarning


class Writer:
    def __init__(self):
        self.strings = []

    def write(self, message):
        self.strings.append(message)


def analyze(code, writer, exceptions_list):
    try:
        result = parse(code)
    except SQFParserError as e:
        writer.write('[%d,%d]:%s\n' % (e.position[0], e.position[1] - 1, e.message))
        exceptions_list += [e]
        return

    exceptions = sqf.analyzer.analyze(result).exceptions
    for e in exceptions:
        writer.write('[%d,%d]:%s\n' % (e.position[0], e.position[1] - 1, e.message))
    exceptions_list += exceptions


def analyze_dir(directory, writer, exceptions_list):
    """
    Analyzes a directory recursively
    """
    for root, dirs, files in os.walk(directory):
        files.sort()
        for file in files:
            if file.endswith(".sqf"):
                file_path = os.path.join(root, file)

                writer_helper = Writer()

                with open(file_path) as f:
                    analyze(f.read(), writer_helper, exceptions_list)

                if writer_helper.strings:
                    writer.write(os.path.relpath(file_path, directory) + '\n')
                    for string in writer_helper.strings:
                        writer.write('\t%s' % string)
    return writer


def readable_dir(prospective_dir):
    if not os.path.isdir(prospective_dir):
        raise Exception("readable_dir:{0} is not a valid path".format(prospective_dir))
    if os.access(prospective_dir, os.R_OK):
        return prospective_dir
    else:
        raise Exception("readable_dir:{0} is not a readable dir".format(prospective_dir))


def parse_args(args):
    parser = argparse.ArgumentParser(description="Static Analyzer of SQF code")
    parser.add_argument('file', nargs='?', type=argparse.FileType('r'), default=None,
                        help='The full path of the file to be analyzed')
    parser.add_argument('-d', '--directory', nargs='?', type=readable_dir, default=None,
                        help='The full path of the directory to recursively analyse sqf files on')
    parser.add_argument('-o', '--output', nargs='?', type=argparse.FileType('w'), default=None,
                        help='File path to redirect the output to (default to stdout)')
    parser.add_argument('-e', '--exit', type=str, default='',
                        help='How the parser should exit. \'\': exit code 0;\n'
                             '\'e\': exit with code 1 when any error is found;\n'
                             '\'w\': exit with code 1 when any error or warning is found.')

    return parser.parse_args(args)


def entry_point(args):
    args = parse_args(args)

    if args.output is None:
        writer = sys.stdout
    else:
        writer = args.output

    exceptions_list = []

    if args.file is None and args.directory is None:
        code = sys.stdin.read()
        analyze(code, writer, exceptions_list)
    elif args.file is not None:
        code = args.file.read()
        args.file.close()
        analyze(code, writer, exceptions_list)
    else:
        analyze_dir(args.directory, writer, exceptions_list)

    if args.output is not None:
        writer.close()

    exit_code = 0
    if args.exit == 'e':
        errors = [e for e in exceptions_list if isinstance(e, SQFParserError)]
        exit_code = int(len(errors) != 0)
    elif args.exit == 'w':
        errors_and_warnings = [e for e in exceptions_list if isinstance(e, (SQFWarning, SQFParserError))]
        exit_code = int(len(errors_and_warnings) != 0)
    return int(exit_code)


def main():
    sys.exit(entry_point(sys.argv[1:]))


if __name__ == "__main__":
    main()



class SQFError(Exception):
    """
    Raised by the parser and analyzer
    """
    def __init__(self, position, message):
        assert(isinstance(position, tuple))
        self.position = position
        self.message = message.replace("\n", "\\n").replace("\t", "\\t").replace("\r", "\\r")


class SQFParserError(SQFError):
    """
    Raised by the parser and analyzer
    """
    def __init__(self, position, message):
        super().__init__(position, "error:%s" % message)


class SQFParenthesisError(SQFParserError):
    pass


class SQFWarning(SQFError):
    """
    Something that the interpreter understands but that is a bad practice or potentially
    semantically incorrect.
    """
    def __init__(self, position, message):
        super().__init__(position, "warning:%s" % message)



from copy import deepcopy

from sqf.types import Statement, Code, Nothing, Variable, Array, String, Type, File, BaseType, \
    Number, Preprocessor, Script, Anything
from sqf.interpreter_types import InterpreterType, PrivateType, ForType, SwitchType, \
    DefineStatement, DefineResult, IfDefResult
from sqf.keywords import Keyword, PREPROCESSORS
from sqf.expressions import UnaryExpression, BinaryExpression
from sqf.exceptions import SQFParserError, SQFWarning
from sqf.base_interpreter import BaseInterpreter
from sqf.database import EXPRESSIONS
from sqf.common_expressions import COMMON_EXPRESSIONS, ForEachExpression, ElseExpression
from sqf.expressions_cache import values_to_expressions, build_database
from sqf.parser_types import Comment
from sqf.parser import parse


def all_equal(iterable):
    return iterable.count(iterable[0]) == len(iterable)


# Replace all expressions in `database` by expressions from `COMMON_EXPRESSIONS` with the same signature
for exp in COMMON_EXPRESSIONS:
    if exp in EXPRESSIONS:
        EXPRESSIONS.remove(exp)
    EXPRESSIONS.append(exp)


EXPRESSIONS_MAP = build_database(EXPRESSIONS)


def is_undefined_define(base_tokens):
    if len(base_tokens) == 2:
        if isinstance(base_tokens[0], Statement):
            first = base_tokens[0].base_tokens[0]
        else:
            first = base_tokens[0]
        return str(first)[0].isupper() and str(base_tokens[1])[0] == '('

    return False


class UnexecutedCode:
    """
    A piece of code that needs to be re-run on a contained env to check for issues.
    We copy the state of the analyzer (namespaces) so we get what that code would run.
    """
    def __init__(self, code, analyzer):
        self.namespaces = deepcopy(analyzer._namespaces)
        self.namespace_name = analyzer.current_namespace.name
        self.code = code
        self.position = code.position
        self.delete_scope_level = analyzer.delete_scope_level


class Analyzer(BaseInterpreter):
    """
    The Analyzer. This is an interpreter that:
    * runs SQF statements that accepts unknown types
    * Stores exceptions instead of rising them.
    * Runs code that is declared but not called.
    """
    COMMENTS_FOR_PRIVATE = {'IGNORE_PRIVATE_WARNING', 'USES_VARIABLES'}

    def __init__(self, all_vars=None):
        super().__init__(all_vars)
        self.exceptions = []

        self.privates = set()
        self.unevaluated_interpreter_tokens = []
        self._unexecuted_codes = {}
        self._executed_codes = {}  # executed code -> result

        self.variable_uses = {}

        # a counter used by `self.assign` to identify if a variable is deleted (assigned to Anything) or not.
        self.delete_scope_level = 0

        # list of variables that we currently know the type during the script.
        self.undefined_variables = set()

    def exception(self, exception):
        self.exceptions.append(exception)

    @staticmethod
    def code_key(code):
        return code.position, str(code)

    @staticmethod
    def exe_code_key(code, extra_scope):
        if extra_scope is None:
            extra_scope = {}
        return str(code), tuple((x, type(extra_scope[x])) for x in sorted(extra_scope.keys()))

    def value(self, token, namespace_name=None):
        """
        Given a single token, recursively evaluates and returns its value
        """
        if namespace_name is None:
            namespace_name = self.current_namespace.name

        assert(isinstance(token, BaseType))
        if isinstance(token, IfDefResult):
            for x in token.result:
                x.set_position(token.position)
                result = self.value(self.execute_token(x))
        elif isinstance(token, DefineResult):
            token.result.set_position(token.position)
            result = self.value(self.execute_token(token.result))
        elif isinstance(token, Statement):
            result = self.value(self.execute_token(token))
        elif isinstance(token, Variable):
            scope = self.get_scope(token.name, namespace_name)
            if scope.level == 0 and not token.is_global:
                self.exception(
                    SQFWarning(token.position, 'Local variable "%s" is not from this scope (not private)' % token))

            try:
                result = scope[token.name]
            except KeyError:
                result = self.private_default_class()
            result.position = token.position

            key = '%s_%s_%s' % (namespace_name, scope.level, scope.normalize(token.name))
            if key in self.variable_uses:
                self.variable_uses[key]['count'] += 1

        elif isinstance(token, Array) and not token.is_undefined:
            result = Array([self.value(self.execute_token(s)) for s in token.value])
            result.position = token.position
        else:
            null_expressions = values_to_expressions([token], EXPRESSIONS_MAP, EXPRESSIONS)
            if null_expressions:
                result = null_expressions[0].execute([token], self)
            else:
                result = token
            result.position = token.position

        if isinstance(result, Code) and self.code_key(result) not in self._unexecuted_codes:
            self._unexecuted_codes[self.code_key(result)] = UnexecutedCode(result, self)

        return result

    def execute_token(self, token):
        """
        Given a single token, recursively evaluate it without returning its value (only type)
        """
        # interpret the statement recursively
        if isinstance(token, Statement):
            result = self.execute_single(statement=token)
            # we do not want the position of the statement, but of the token, so we do not
            # store it here
        elif isinstance(token, Array) and token.value is not None:
            result = Array([self.execute_token(s) for s in token.value])
            result.position = token.position
        else:
            result = token
            result.position = token.position

        return result

    def execute_unexecuted_code(self, code_key, extra_scope=None, own_namespace=False):
        """
        Executes a code in a dedicated env and put consequence exceptions in self.

        own_namespace: whether the execution uses the current local variables or no variables
        """
        container = self._unexecuted_codes[code_key]

        analyzer = Analyzer()
        if not own_namespace:
            analyzer._namespaces = container.namespaces
        analyzer.variable_uses = self.variable_uses
        analyzer.delete_scope_level = container.delete_scope_level

        file = File(container.code._tokens)
        file.position = container.position

        this = Anything()
        this.position = container.position

        analyzer.execute_code(file, extra_scope=extra_scope,
                              namespace_name=container.namespace_name, delete_mode=True)

        self.exceptions.extend(analyzer.exceptions)

    def execute_code(self, code, extra_scope=None, namespace_name='missionnamespace', delete_mode=False):
        key = self.code_key(code)
        exe_code_key = self.exe_code_key(code, extra_scope)

        if key in self._unexecuted_codes:
            del self._unexecuted_codes[key]
        if exe_code_key in self._executed_codes:
            outcome = self._executed_codes[exe_code_key]
        else:
            self.delete_scope_level += delete_mode
            outcome = super().execute_code(code, extra_scope, namespace_name)
            self.delete_scope_level -= delete_mode
            self._executed_codes[exe_code_key] = outcome

        if isinstance(code, File):
            for key in self._unexecuted_codes:
                self.execute_unexecuted_code(key)

            # collect `private` statements that have a variable but were not collected by the assignment operator
            # this check is made at the scope level
            for private in self.privates:
                self.exception(SQFWarning(private.position, 'private argument must be a string.'))

            # this check is made at the scope level
            for token in self.unevaluated_interpreter_tokens:
                self.exception(SQFWarning(token.position, 'helper type "%s" not evaluated' % token.__class__.__name__))

            # this check is made at script level
            if not delete_mode:
                # collect variables that were not used
                for key in self.variable_uses:
                    if self.variable_uses[key]['count'] == 0:
                        variable = self.variable_uses[key]['variable']
                        self.exception(
                            SQFWarning(variable.position, 'Variable "%s" not used' % variable.value))

        return outcome

    def _parse_params_args(self, arguments, base_token):
        if isinstance(arguments, Anything) or (isinstance(arguments, Array) and arguments.is_undefined):
            return [Anything() for _ in range(len(base_token))]
        return super()._parse_params_args(arguments, base_token)

    def _add_private(self, variable):
        super()._add_private(variable)
        scope = self.current_scope
        key = '%s_%s_%s' % (self.current_namespace.name, scope.level, scope.normalize(variable.value))
        self.variable_uses[key] = {'count': 0, 'variable': variable}

    def assign(self, lhs, rhs_v):
        """
        Assigns the rhs_v to the lhs variable.
        """
        lhs_name = lhs.name
        lhs_position = lhs.position

        scope = self.get_scope(lhs_name)

        try:
            lhs_t = type(scope[lhs.name])
        except KeyError:
            lhs_t = self.private_default_class
        rhs_t = type(rhs_v)

        if scope.level == 0:
            # global variable becomes undefined when:
            # 1. it changes type AND
            # 2. it is modified on a higher delete scope (e.g. if {}) or it already has a defined type
            if (lhs_t != Anything or self.delete_scope_level > scope.level) and \
                    lhs_t != rhs_t and lhs_name not in self.undefined_variables:
                self.undefined_variables.add(lhs_name)

        if scope.level == 0:
            if lhs_name in self.undefined_variables:
                rhs_t = Anything
        elif lhs_t != rhs_t and self.delete_scope_level >= scope.level:
            rhs_t = Anything

        scope[lhs_name] = rhs_t()

        if scope.level == 0 and lhs_name.startswith('_'):
            self.exception(
                SQFWarning(lhs_position, 'Local variable "%s" assigned to an outer scope (not private)' % lhs_name))

    def execute_single(self, statement):
        assert(isinstance(statement, Statement))

        outcome = Nothing()
        outcome.position = statement.position

        base_tokens = []
        for token in statement.tokens:
            if not statement.is_base_token(token):
                self.execute_other(token)
            else:
                base_tokens.append(token)

        if not base_tokens:
            return outcome

        # operations that cannot evaluate the value of all base_tokens
        if type(base_tokens[0]) == DefineStatement:
            return base_tokens[0]
        elif base_tokens[0] == Preprocessor("#include"):
            if len(base_tokens) != 2:
                exception = SQFParserError(base_tokens[0].position, "#include requires one argument")
                self.exception(exception)
            elif type(self.execute_token(base_tokens[1])) != String:
                exception = SQFParserError(base_tokens[0].position, "#include first argument must be a string")
                self.exception(exception)
            return outcome
        elif isinstance(base_tokens[0], Keyword) and base_tokens[0].value in PREPROCESSORS:
            # remaining preprocessors are ignored
            return outcome
        elif len(base_tokens) == 2 and base_tokens[0] == Keyword('private'):
            # the rhs may be a variable, so we cannot get the value
            rhs = self.execute_token(base_tokens[1])
            if isinstance(rhs, String):
                self.add_privates([rhs])
            elif isinstance(rhs, Array):
                value = self.value(rhs)
                if value.is_undefined:
                    self.exception(SQFWarning(base_tokens[0].position,
                                              'Obfuscated statement. Consider explicitly set what is private.'))
                else:
                    self.add_privates(value)
            elif isinstance(rhs, Variable):
                var = String('"' + rhs.name + '"')
                var.position = rhs.position
                self.add_privates([var])
                outcome = PrivateType(rhs)
                outcome.position = rhs.position
                self.privates.add(outcome)
            else:
                self.exception(SQFParserError(base_tokens[0].position, '`private` used incorrectly'))
            return outcome
        # assignment operator
        elif len(base_tokens) == 3 and base_tokens[1] == Keyword('='):
            lhs = self.execute_token(base_tokens[0])
            if isinstance(lhs, PrivateType):
                self.privates.remove(lhs)
                lhs = lhs.variable
            else:
                lhs = self.get_variable(base_tokens[0])

            if not isinstance(lhs, Variable):
                self.exception(SQFParserError(base_tokens[0].position, 'lhs of assignment operator must be a variable'))
            else:
                # if the rhs_v is code and calls `lhs` (recursion) it will assume lhs is anything (and not Nothing)
                scope = self.get_scope(lhs.name)
                if lhs.name not in scope or isinstance(scope[lhs.name], Nothing):
                    scope[lhs.name] = Anything()

                rhs_v = self.value(base_tokens[2])
                self.assign(lhs, rhs_v)
                if not statement.ending:
                    outcome = rhs_v
            return outcome
        # A variable can only be evaluated if we need its value, so we will not call its value until the very end.
        elif len(base_tokens) == 1 and type(base_tokens[0]) in (Variable, Array):
            return self.execute_token(base_tokens[0])
        # heuristic for defines (that are thus syntactically correct):
        #   - is keyword but upper cased
        #   - first token string starts uppercased
        elif len(base_tokens) == 1 and type(base_tokens[0]) == Keyword and str(base_tokens[0])[0].isupper():
            outcome = Variable(str(base_tokens[0]))
            outcome.position = base_tokens[0].position
            return outcome
        elif is_undefined_define(base_tokens):
            # get all arguments and compute their value to analyze them
            if isinstance(base_tokens[1].base_tokens[0], Statement):
                sub_tokens = base_tokens[1].base_tokens[0].base_tokens
            else:
                sub_tokens = base_tokens[0]
            for sub_token in sub_tokens:
                self.value(sub_token)

            # finally, build the outcome
            outcome = Anything()
            outcome.position = base_tokens[0].position
            return outcome

        # evaluate all the base_tokens, trying to obtain their values
        values = []
        tokens = []
        for token in base_tokens:
            t = self.execute_token(token)
            v = self.value(t)
            tokens.append(t)
            values.append(v)

        # try to find a match for any expression, both typed and un-typed
        case_found = None
        possible_expressions = values_to_expressions(values, EXPRESSIONS_MAP, EXPRESSIONS)
        for case in possible_expressions:
            if case.is_signature_match(values):  # match first occurrence
                case_found = case
                break

        if case_found:
            # if exact match, we run the expression.
            if case_found.is_match(values):
                # parse and execute the string that is code (to count usage of variables)
                if case_found.keyword == Keyword('isnil') and type(values[1]) == String or \
                   case_found.keyword == Keyword('configClasses'):
                    code_position = {'isnil': 1, 'configclasses': 0}[case_found.keyword.unique_token]
                    extra_scope = {'isnil': None, 'configclasses': {'_x': Anything()}}[case_found.keyword.unique_token]

                    # when the string is undefined, there is no need to evaluate it.
                    if not values[code_position].is_undefined:
                        try:
                            code = Code([parse(values[code_position].value)])
                            code.position = values[code_position].position
                            self.execute_code(code, extra_scope=extra_scope)
                        except SQFParserError as e:
                            self.exceptions.append(
                                SQFParserError(values[code_position].position,
                                               'Error while parsing a string to code: %s' % e.message))
                # finally, execute the statement
                outcome = case_found.execute(values, self)
            elif len(possible_expressions) == 1 or all_equal([x.return_type for x in possible_expressions]):
                return_type = possible_expressions[0].return_type
                if isinstance(case_found, (ForEachExpression, ElseExpression)):
                    outcome = Anything()
                elif return_type is not None:
                    outcome = return_type()
                if return_type == ForType:
                    outcome.copy(values[0])
                elif case_found.keyword == Keyword('call'):
                    outcome = Anything()
            else:
                # when a case is found but we cannot decide on the type, it is anything
                outcome = Anything()

            extra_scope = None
            if case_found.keyword in (Keyword('select'), Keyword('apply'), Keyword('count')):
                extra_scope = {'_x': Anything()}
            elif case_found.keyword == Keyword('foreach'):
                extra_scope = {'_foreachindex': Number(), '_x': Anything()}
            elif case_found.keyword == Keyword('catch'):
                extra_scope = {'_exception': Anything()}
            elif case_found.keyword == Keyword('spawn'):
                extra_scope = {'_thisScript': Script()}
            elif case_found.keyword == Keyword('do') and type(values[0]) == ForType:
                extra_scope = {values[0].variable.value: Number()}
            for value, t_or_v in zip(values, case_found.types_or_values):
                # execute all pieces of code
                if t_or_v == Code and isinstance(value, Code) and self.code_key(value) not in self._executed_codes:
                    if case_found.keyword == Keyword('spawn'):
                        self.execute_unexecuted_code(self.code_key(value), extra_scope, True)
                        # this code was executed, so it does not need to be evaluated on an un-executed env.
                        del self._unexecuted_codes[self.code_key(value)]
                    else:
                        self.execute_code(value, extra_scope=extra_scope, namespace_name=self.current_namespace.name, delete_mode=True)

                # remove evaluated interpreter tokens
                if isinstance(value, InterpreterType) and value in self.unevaluated_interpreter_tokens:
                    self.unevaluated_interpreter_tokens.remove(value)

            assert(isinstance(outcome, Type))
        elif len(values) == 1:
            if not isinstance(values[0], Type):
                self.exception(
                    SQFParserError(statement.position, '"%s" is syntactically incorrect (missing ;?)' % statement))
            outcome = values[0]
        elif isinstance(base_tokens[0], Variable) and base_tokens[0].is_global:
            # statements starting with a global are likely defined somewhere else
            # todo: catch globals with statements and without statements
            pass
        elif len(possible_expressions) > 0:
            if isinstance(possible_expressions[0], UnaryExpression):
                types_or_values = []
                for exp in possible_expressions:
                    types_or_values.append(exp.types_or_values[1].__name__)

                keyword_name = possible_expressions[0].types_or_values[0].value

                message = 'Unary operator "%s" only accepts argument of types [%s] (rhs is %s)' % \
                          (keyword_name, ','.join(types_or_values), values[1].__class__.__name__)
            elif isinstance(possible_expressions[0], BinaryExpression):
                types_or_values = []
                for exp in possible_expressions:
                    types_or_values.append('(%s,%s)' % (exp.types_or_values[0].__name__, exp.types_or_values[2].__name__))

                keyword_name = possible_expressions[0].types_or_values[1].value

                message = 'Binary operator "{0}" arguments must be [{1}]'.format(
                    keyword_name, ','.join(types_or_values))
                if values[0].__class__.__name__ not in [x[0] for x in types_or_values]:
                    message += ' (lhs is %s' % values[0].__class__.__name__
                if values[0].__class__.__name__ not in [x[1] for x in types_or_values]:
                    message += ', rhs is %s)' % values[2].__class__.__name__
                else:
                    message += ')'
            else:
                assert False

            self.exception(SQFParserError(values[1].position, message))
            # so the error does not propagate further
            outcome = Anything()
            outcome.position = base_tokens[0].position
        else:
            helper = ' '.join(['<%s(%s)>' % (type(t).__name__, t) for t in tokens])
            self.exception(
                SQFParserError(base_tokens[-1].position, 'can\'t interpret statement (missing ;?): %s' % helper))
            # so the error does not propagate further
            outcome = Anything()
            outcome.position = base_tokens[0].position

        if isinstance(outcome, InterpreterType) and \
            outcome not in self.unevaluated_interpreter_tokens and type(outcome) not in (SwitchType, PrivateType, DefineStatement):
            # switch type can be not evaluated, e.g. for `case A; case B: {}`
            self.unevaluated_interpreter_tokens.append(outcome)

        assert(isinstance(outcome, BaseType))
        # the position of Private is different because it can be passed from analyzer to analyzer,
        # and we want to keep the position of the outermost analyzer.
        if not isinstance(outcome, PrivateType):
            outcome.position = base_tokens[0].position

        if statement.ending:
            outcome = Nothing()
            outcome.position = base_tokens[0].position

        return outcome

    def execute_other(self, statement):
        if isinstance(statement, Comment):
            string = str(statement)[2:]
            matches = [x for x in self.COMMENTS_FOR_PRIVATE if string.startswith(x)]
            if matches:
                length = len(matches[0]) + 1  # +1 for the space
                try:
                    parsed_statement = parse(string[length:])
                    array = parsed_statement[0][0]
                    assert(isinstance(array, Array))
                    self.add_privates(self.value(array))
                    # these are unknown values.
                    for token in array.value:
                        if isinstance(token, Statement):
                            token = token.base_tokens[0]
                        self.current_scope[token.value] = Anything()
                except Exception:
                    self.exception(SQFWarning(statement.position, '{0} comment must be `//{0} ["var1",...]`'.format(matches[0])))


def analyze(statement, analyzer=None):
    assert (isinstance(statement, Statement))
    if analyzer is None:
        analyzer = Analyzer()

    file = File(statement.tokens)

    file.position = (1, 1)

    arg = Anything()
    arg.position = (1, 1)

    analyzer.execute_code(file, extra_scope={'_this': arg})

    return analyzer



from sqf.base_type import ParserType


class Comment(ParserType):

    def __init__(self, string):
        super().__init__()
        assert (string.startswith('/*') or string.startswith('//'))
        self._string = string

    def __str__(self):
        return self._string

    def __repr__(self):
        return ('C(%s)' % self).replace('\r\n', r'\r\n').replace('\n', r'\n')


class Space(ParserType):
    def __str__(self):
        return ' '

    def __repr__(self):
        return '\' \''


class Tab(ParserType):
    def __str__(self):
        return '\t'

    def __repr__(self):
        return '\\t'


class EndOfLine(ParserType):
    def __init__(self, value):
        super().__init__()
        assert(value in ['\n', '\r\n'])
        self.value = value

    def __str__(self):
        return self.value

    def __repr__(self):
        return '<EOL>'


class BrokenEndOfLine(ParserType):
    def __str__(self):
        return '\\\n'

    def __repr__(self):
        return '<\EOL>'


class EndOfFile(ParserType):
    def __str__(self):
        return ''

    def __repr__(self):
        return '<\EOF>'


class ParserKeyword(ParserType):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value



from sqf.types import Keyword, Nothing, Anything, Type
from sqf.interpreter_types import InterpreterType


class Expression:
    """
    A generic class to represent an expression. The expression matches according to the
    types of their elements, listed in `types`.
    """
    def __init__(self, types_or_values, return_type):
        self.types_or_values = tuple(types_or_values)
        self.return_type = return_type
        for t_or_v in self.types_or_values:
            assert (isinstance(t_or_v, (Type, Keyword)) or issubclass(t_or_v, Type))
        assert(return_type is None or issubclass(return_type, Type))

    def is_match(self, values, exact=True):
        """
        Given a list of values, returns a list of matches when the values
        match or not each condition of the expression
        """
        if len(values) != len(self.types_or_values):
            return False

        for i, (t_or_v, value) in enumerate(zip(self.types_or_values, values)):
            if isinstance(t_or_v, (Type, Keyword)):  # it is a value
                if value != t_or_v:
                    return False
            else:  # it is a type
                if not (isinstance(value, t_or_v) or
                            (not exact and type(value) == Anything and
                             not issubclass(t_or_v, InterpreterType))):
                    return False
        return True

    def is_signature_match(self, values):
        return self.is_match(values, exact=False)

    def execute(self, values, interpreter):
        raise NotImplementedError

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.types_or_values)

    def __eq__(self, other):
        if issubclass(other.__class__, Expression):
            return self.types_or_values == other.types_or_values
        else:
            return False

    @property
    def keyword(self):
        raise NotImplementedError

    def _result_to_typed_result(self, value):
        if self.return_type is None:
            return value
        elif self.return_type in (Anything, Nothing):
            return self.return_type()
        else:
            if isinstance(value, tuple):
                return self.return_type(*value)
            else:
                return self.return_type(value)


class UnaryExpression(Expression):
    def __init__(self, op, rhs_type, return_type, action=None):
        assert (isinstance(op, Keyword))
        super().__init__([op, rhs_type], return_type)
        if action is None and return_type is None:
            action = lambda rhs, i: i.private_default_class()
        elif action is None:
            action = lambda rhs, i: None
        self.action = action

    def execute(self, values, interpreter):
        result = self.action(values[1], interpreter)
        return self._result_to_typed_result(result)

    @property
    def keyword(self):
        return self.types_or_values[0]


class BinaryExpression(Expression):
    def __init__(self, lhs_type, op, rhs_type, return_type, action=None):
        assert(isinstance(op, Keyword))
        super().__init__([lhs_type, op, rhs_type], return_type)
        if action is None and return_type is None:
            action = lambda lhs, rhs, i: i.private_default_class()
        elif action is None:
            action = lambda lhs, rhs, i: None
        self.action = action

    def execute(self, values, interpreter):
        result = self.action(values[0], values[2], interpreter)
        return self._result_to_typed_result(result)

    @property
    def keyword(self):
        return self.types_or_values[1]


class NullExpression(Expression):
    def __init__(self, op, return_type, action=None):
        assert(isinstance(op, Keyword))
        assert(return_type is not None)
        super().__init__([op], return_type)
        if action is None:
            action = lambda i: None
        self.action = action

    def execute(self, values, interpreter):
        result = self.action(interpreter)
        return self._result_to_typed_result(result)

    @property
    def keyword(self):
        return self.types_or_values[0]



from sqf.types import Keyword
from sqf.expressions import BinaryExpression, UnaryExpression
from sqf.database import EXPRESSIONS

# keywords that are not commands, but part of the language
KEYWORDS = {'=', '\\'}

PREPROCESSORS_UNARY = {'#ifdef', '#ifndef', '#undef', '#include'}
PREPROCESSORS_NULLARY = {'#else', '#endif'}
PREPROCESSORS = {'#define'}.union(PREPROCESSORS_UNARY).union(PREPROCESSORS_NULLARY)

KEYWORDS = KEYWORDS.union(PREPROCESSORS)


NULARY_OPERATORS = set()
UNARY_OPERATORS = set()
BINARY_OPERATORS = set()

for expression in EXPRESSIONS:
    if isinstance(expression, BinaryExpression):
        op = expression.types_or_values[1]
        BINARY_OPERATORS.add(op.value.lower())
    elif isinstance(expression, UnaryExpression):
        op = expression.types_or_values[0]
        UNARY_OPERATORS.add(op.value.lower())
    else:
        op = expression.types_or_values[0]
        NULARY_OPERATORS.add(op)

    KEYWORDS.add(op.value.lower())


OP_ARITHMETIC = [Keyword(s) for s in ('+', '-', '*', '/', '%', 'mod', '^', 'max', 'floor')]

OP_LOGICAL = [Keyword(s) for s in ('&&', 'and', '||', 'or')]

OP_COMPARISON = [Keyword(s) for s in ('==', 'isequalto', '!=', '<', '>', '<=', '>=', '>>')]

NAMESPACES = {'missionnamespace', 'profilenamespace', 'uinamespace', 'parsingnamespace'}

# namespaces are parsed as such
KEYWORDS = KEYWORDS - NAMESPACES

UNARY_OPERATORS.union({'#ifdef', '#ifndef', '#include'})



from sqf.types import Statement, Code, Nothing, Anything, Variable, Array, String, Type, File
from sqf.keywords import Keyword
from sqf.exceptions import SQFParserError, SQFWarning
import sqf.namespace


class BaseInterpreter:
    """
    Base Interpreter used by the analyzer and interpreter
    """
    private_default_class = Anything

    def __init__(self, all_vars=None):
        self._namespaces = {
            'uinamespace': sqf.namespace.Namespace('uinamespace'),
            'parsingnamespace': sqf.namespace.Namespace('parsingnamespace'),
            'missionnamespace': sqf.namespace.Namespace('missionnamespace', all_vars),
            'profilenamespace': sqf.namespace.Namespace('profilenamespace')
        }

        self.current_namespace = self.namespace('missionnamespace')

    def exception(self, exception):
        """
        We can overwrite this method to handle exceptions differently
        """
        raise exception

    def set_global_variable(self, var_name, value):
        assert(isinstance(value, Type))
        self.namespace('missionNamespace').base_scope[var_name] = value

    def namespace(self, name):
        return self._namespaces[name.lower()]

    @property
    def current_scope(self):
        return self.current_namespace.current_scope

    def __getitem__(self, name):
        try:
            return self.get_scope(name)[name]
        except KeyError:
            return self.private_default_class()

    def __contains__(self, name):
        return name in self.get_scope(name)

    def _parse_params_args(self, arguments, base_token):
        """
        How param argument should be interpreted when it is not an array nor Nothing
        """
        return [arguments]

    def _add_params(self, token):
        self.add_privates([token[0]])

    def add_params(self, base_token, arguments=None):
        assert (isinstance(base_token, Array))

        if arguments is None or isinstance(arguments, Nothing):
            arguments = self['_this']
        if isinstance(arguments, Array) and not arguments.is_undefined:
            arguments = arguments.value
        else:
            arguments = self._parse_params_args(arguments, base_token)

        if len(arguments) > len(base_token):
            self.exception(
                SQFWarning(base_token.position,
                           '`params` lhs (%d elements) is larger than rhs (%d elements).'
                           ' Some arguments are ignored.' % (len(arguments), len(base_token))))

        for i, token in enumerate(base_token):
            if isinstance(token, String):
                if token.value == '':
                    continue
                self.add_privates([token])
                if i >= len(arguments):
                    self.exception(SQFWarning(token.position,
                                   '`params` mandatory argument %s is missing in rhs' % token))
                else:
                    self.current_scope[token.value] = arguments[i]
            elif isinstance(token, Array):
                if len(token) in (2, 3, 4):
                    if i < len(arguments) and not isinstance(arguments[i], Nothing):
                        argument = arguments[i]
                    else:
                        argument = token[1]

                    self._add_params(token)
                    self.current_scope[token[0].value] = argument
                else:
                    self.exception(
                        SQFParserError(base_token.position, '`params` array element must have 2-4 elements'))
            else:
                self.exception(SQFParserError(base_token.position, '`params` array element must be a string or array'))
        return True

    def value(self, token, namespace_name=None):
        if isinstance(token, Statement):
            return self.value(self.execute_single(statement=token))
        elif isinstance(token, Variable):
            scope = self.get_scope(token.name, namespace_name)
            try:
                value = scope[token.name]
            except KeyError:
                value = self.private_default_class()
            assert isinstance(value, Type)
            return value
        elif isinstance(token, (Type, Keyword)):
            return token
        else:
            raise NotImplementedError(repr(token))

    def get_variable(self, token):
        if isinstance(token, Statement):
            return self.get_variable(token.base_tokens[0])
        else:
            if not isinstance(token, Variable):
                return Nothing()
            return token

    def execute_token(self, token):
        """
        Given a single token, recursively evaluate it and return its value.
        """
        raise NotImplementedError

    def get_scope(self, name, namespace_name=None):
        if namespace_name is None:
            namespace = self.current_namespace
        else:
            namespace = self.namespace(namespace_name)
        return namespace.get_scope(name)

    def _add_private(self, variable):
        assert isinstance(variable, String)
        self.current_scope[variable.value] = Nothing()

    def add_privates(self, variables):
        """
        Privatizes a list of variables by initializing them on the scope (as Nothing).
        """
        for variable in variables:
            if not isinstance(variable, String):
                self.exception(SQFParserError(variable.position, 'Variable in private must be a string (is %s)' % type(variable)))
                continue

            if not variable.value.startswith('_'):
                self.exception(SQFParserError(variable.position, 'Cannot make global variable "%s" private (underscore missing?)' % variable.value))
                continue
            self._add_private(variable)

    def execute_other(self, statement):
        pass

    def execute_code(self, code, extra_scope=None, namespace_name='missionnamespace'):
        assert (isinstance(code, Code))

        # store the old namespace
        _previous_namespace = self.current_namespace

        # store the executing namespace
        namespace = self.namespace(namespace_name)
        # change to the executing namespace
        self.current_namespace = namespace

        if extra_scope is None:
            extra_scope = {}
        namespace.add_scope(extra_scope)

        # execute the code
        outcome = self.private_default_class()
        outcome.position = code.position
        for statement in code.base_tokens:
            token = self.execute_token(statement)
            if isinstance(token, tuple):
                token = token[0]
            outcome = self.value(token)

        # cleanup
        if not isinstance(code, File):  # so we have access to its scope
            # this has to be the executing namespace because "self.current_namespace" may change
            namespace.del_scope()
        self.current_namespace = _previous_namespace
        return outcome

    def execute_single(self, statement):
        """
        Executes a statement
        """
        raise NotImplementedError



class Scope:
    """
    A scope is a dictionary that stores variables. Its level is controlled by a namespace
    and has no function to the scope itself.
    The values are case insensitive because SQF variables are case-insensitive.
    """
    def __init__(self, level, values=None):
        if values is None:
            values = {}
        self.values = {self.normalize(key): values[key] for key in values}
        self.level = level

    def __contains__(self, name):
        return self.normalize(name) in self.values

    def __getitem__(self, name):
        return self.values[self.normalize(name)]

    def __setitem__(self, name, value):
        self.values[self.normalize(name)] = value

    @staticmethod
    def normalize(name):
        return name.lower()


class Namespace:
    def __init__(self, name, all_vars=None):
        self._stack = [Scope(0, all_vars)]
        self.name = name

    def __repr__(self):
        return '<Namespace %s>' % self.name

    def __getitem__(self, name):
        return self.get_scope(name)[name]

    def __contains__(self, name):
        return name.lower() in self.get_scope(name)

    @property
    def current_scope(self):
        return self._stack[-1]

    @property
    def base_scope(self):
        return self._stack[0]

    def get_scope(self, name):
        if name.startswith('_'):
            for i in reversed(range(1, len(self._stack))):
                scope = self._stack[i]
                if name in scope:
                    return scope
            return self._stack[0]
        else:
            return self._stack[0]

    def add_scope(self, values=None):
        self._stack.append(Scope(len(self._stack), values))

    def del_scope(self):
        del self._stack[-1]



from sqf.types import Statement, Code, Number, Boolean, Nothing, Variable, Array, String, Type, File
from sqf.interpreter_types import PrivateType
from sqf.keywords import Keyword
from sqf.parser import parse
from sqf.exceptions import SQFParserError
from sqf.common_expressions import COMMON_EXPRESSIONS as EXPRESSIONS
from sqf.interpreter_expressions import INTERPRETER_EXPRESSIONS
from sqf.base_interpreter import BaseInterpreter


# Replace all expressions in `database` by expressions from `COMMON_EXPRESSIONS` with the same signature
for exp in INTERPRETER_EXPRESSIONS:
    if exp in EXPRESSIONS:
        EXPRESSIONS.remove(exp)
    EXPRESSIONS.append(exp)


class Interpreter(BaseInterpreter):
    private_default_class = Nothing

    def __init__(self, all_vars=None):
        super().__init__(all_vars)

        self._simulation = None
        self._client = None

    @property
    def simulation(self):
        return self._simulation

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client):
        self._client = client
        self._simulation = client.simulation

    def _add_params(self, token):
        super()._add_params(token)
        lhs = token[0].value
        scope = self.get_scope(lhs)
        scope[lhs] = token[1]

    def execute_token(self, token):
        """
        Given a single token, recursively evaluate it and return its value.
        """
        # interpret the statement recursively
        if isinstance(token, Statement):
            result = self.execute_single(statement=token)
        elif isinstance(token, Array):
            # empty statements are ignored
            result = Array([self.execute_token(s)[1] for s in token.value if s])
        elif token == Keyword('isServer'):
            result = Boolean(self.client.is_server)
        elif token == Keyword('isDedicated'):
            result = Boolean(self.client.is_dedicated)
        else:
            result = token

        result.position = token.position
        return result, self.value(result)

    def execute_single(self, statement):
        assert(not isinstance(statement, Code))

        outcome = Nothing()
        _outcome = outcome

        # evaluate the types of all tokens
        base_tokens = statement.base_tokens
        values = []
        tokens = []
        types = []

        for token in base_tokens:
            t, v = self.execute_token(token)
            values.append(v)
            tokens.append(t)
            types.append(type(v))

        case_found = None
        for case in EXPRESSIONS:
            if case.is_match(values):
                case_found = case
                break

        if case_found is not None:
            outcome = case_found.execute(values, self)
        # todo: replace all elif below by expressions
        elif len(tokens) == 2 and tokens[0] == Keyword('publicVariable'):
            if not isinstance(tokens[1], String) or tokens[1].value.startswith('_'):
                raise SQFParserError(statement.position, 'Interpretation of "%s" failed' % statement)

            var_name = tokens[1].value
            scope = self.get_scope(var_name, 'missionNamespace')
            self.simulation.broadcast(var_name, scope[var_name])

        elif len(tokens) == 2 and tokens[0] == Keyword('publicVariableServer'):
            if not isinstance(tokens[1], String) or tokens[1].value.startswith('_'):
                raise SQFParserError(statement.position, 'Interpretation of "%s" failed' % statement)

            var_name = tokens[1].value
            scope = self.get_scope(var_name, 'missionNamespace')
            self.simulation.broadcast(var_name, scope[var_name], -1)  # -1 => to server

        elif len(tokens) == 2 and tokens[0] == Keyword('private'):
            if isinstance(values[1], String):
                self.add_privates([values[1]])
            elif isinstance(values[1], Array):
                self.add_privates(values[1].value)
            elif isinstance(base_tokens[1], Statement) and isinstance(base_tokens[1].base_tokens[0], Variable):
                var = base_tokens[1].base_tokens[0]
                self.add_privates([String('"' + var.name + '"')])
                outcome = PrivateType(var)
        # binary operators
        elif len(tokens) == 3 and tokens[1] in (Keyword('='), Keyword('publicVariableClient')):
            # it is a binary statement: token, operation, token
            lhs = tokens[0]
            lhs_v = values[0]
            lhs_t = types[0]

            op = tokens[1]
            rhs = tokens[2]
            rhs_v = values[2]
            rhs_t = types[2]

            if op == Keyword('='):
                if isinstance(lhs, PrivateType):
                    lhs = lhs.variable
                else:
                    lhs = self.get_variable(base_tokens[0])

                if not isinstance(lhs, Variable) or not isinstance(rhs_v, Type):
                    raise SQFParserError(statement.position, 'Interpretation of "%s" failed' % statement)

                scope = self.get_scope(lhs.name)
                scope[lhs.name] = rhs_v
                outcome = rhs
            elif op == Keyword('publicVariableClient'):
                if not lhs_t == Number or rhs.value.startswith('_'):
                    raise SQFParserError(statement.position, 'Interpretation of "%s" failed' % statement)
                client_id = lhs.value
                var_name = rhs.value
                scope = self.get_scope(var_name, 'missionNamespace')
                self.simulation.broadcast(var_name, scope[var_name], client_id)
        # code, variables and values
        elif len(tokens) == 1 and isinstance(tokens[0], (Type, Keyword)):
            outcome = values[0]
        else:
            raise SQFParserError(statement.position, 'Interpretation of "%s" failed' % statement)

        if statement.ending:
            outcome = _outcome
        assert(type(outcome) in (Nothing,Keyword) or not outcome.is_undefined)
        return outcome


def interpret(script, interpreter=None):
    if interpreter is None:
        interpreter = Interpreter()
    assert(isinstance(interpreter, Interpreter))

    statements = parse(script)

    file = File(statements._tokens)
    file.position = (1, 1)

    outcome = interpreter.execute_code(file, extra_scope={'_this': Nothing()})

    return interpreter, outcome



from collections import defaultdict
import re

import sqf.base_type
from sqf.base_tokenizer import tokenize

from sqf.exceptions import SQFParenthesisError, SQFParserError
from sqf.types import Statement, Code, Number, Boolean, Variable, Array, String, Keyword, Namespace, Preprocessor, ParserType
from sqf.keywords import KEYWORDS, NAMESPACES, PREPROCESSORS
from sqf.parser_types import Comment, Space, Tab, EndOfLine, BrokenEndOfLine, EndOfFile, ParserKeyword
from sqf.interpreter_types import DefineStatement, DefineResult, IfDefStatement, IfDefResult
from sqf.parser_exp import parse_exp


def rindex(the_list, value):
    return len(the_list) - the_list[::-1].index(value) - 1


_LEVELS = {'[]': 0, '()': 0, '{}': 0, '#include': 0, '#define': 0, 'ifdef': 0, 'ifdef_open_close': 0}


STOP_KEYWORDS = {
    'single': (ParserKeyword(';'),),
    'both': (ParserKeyword(';'), ParserKeyword(',')),
}

OPEN_PARENTHESIS = (ParserKeyword('['), ParserKeyword('('), ParserKeyword('{'))
CLOSE_PARENTHESIS = (ParserKeyword(']'), ParserKeyword(')'), ParserKeyword('}'))


def get_coord(tokens):
    return sqf.base_type.get_coord(''.join([str(x) for x in tokens]))


def add_coords(coord1, tokens):
    coord2 = get_coord(tokens)
    return coord1[0] + coord2[0], coord1[1] + coord2[1] - 1


def identify_token(token):
    """
    The function that converts a token from tokenize to a BaseType.
    """
    if isinstance(token, (Comment, String)):
        return token
    if token == ' ':
        return Space()
    if token == '\t':
        return Tab()
    if token == '\\\n':
        return BrokenEndOfLine()
    if token in ('(', ')', '[', ']', '{', '}', ',', ';'):
        return ParserKeyword(token)
    if token in ('\n', '\r\n'):
        return EndOfLine(token)
    if token in ('true', 'false'):
        return Boolean(token == 'true')
    try:
        return Number(int(token))
    except ValueError:
        pass
    try:
        return Number(float(token))
    except ValueError:
        pass
    if token in PREPROCESSORS:
        return Preprocessor(token)
    if token.lower() in NAMESPACES:
        return Namespace(token)
    elif token.lower() in KEYWORDS:
        return Keyword(token)
    else:
        return Variable(token)


def replace_in_expression(expression, args, arg_indexes, all_tokens):
    """
    Recursively replaces matches of `args` in expression (a list of Types).
    """
    replacing_expression = []
    for token in expression:
        if isinstance(token, Statement):
            new_expression = replace_in_expression(token.content, args, arg_indexes, all_tokens)
            token = Statement(new_expression, ending=token.ending, parenthesis=token.parenthesis)
            replacing_expression.append(token)
        else:
            for arg, arg_index in zip(args, arg_indexes):
                if str(token) == arg:
                    replacing_expression.append(all_tokens[arg_index])
                    break
            else:
                replacing_expression.append(token)
    return replacing_expression


def parse_strings_and_comments(all_tokens):
    """
    Function that parses the strings of a script, transforming them into `String`.
    """
    string = ''  # the buffer for the activated mode
    tokens = []  # the final result
    in_double = False
    mode = None  # [None, "string_single", "string_double", "comment_line", "comment_bulk"]

    for i, token in enumerate(all_tokens):
        if mode == "string_double":
            string += token
            if token == '"':
                if in_double:
                    in_double = False
                elif not in_double and i != len(all_tokens) - 1 and all_tokens[i+1] == '"':
                    in_double = True
                else:
                    tokens.append(String(string))
                    mode = None
                    in_double = False
        elif mode == "string_single":
            string += token
            if token == "'":
                if in_double:
                    in_double = False
                elif not in_double and i != len(all_tokens) - 1 and all_tokens[i + 1] == "'":
                    in_double = True
                else:
                    tokens.append(String(string))
                    mode = None
                    in_double = False
        elif mode == "comment_bulk":
            string += token
            if token == '*/':
                mode = None
                tokens.append(Comment(string))
                string = ''
        elif mode == "comment_line":
            string += token
            if token in ('\n', '\r\n'):
                mode = None
                tokens.append(Comment(string))
                string = ''
        else:  # mode is None
            if token == '"':
                string = token
                mode = "string_double"
            elif token == "'":
                string = token
                mode = "string_single"
            elif token == '/*':
                string = token
                mode = "comment_bulk"
            elif token == '//':
                string = token
                mode = "comment_line"
            else:
                tokens.append(token)

    if mode in ("comment_line", "comment_bulk"):
        tokens.append(Comment(string))
    elif mode is not None:
        raise SQFParserError(get_coord(tokens), 'String is not closed')

    return tokens


def _analyze_simple(tokens):
    return Statement(tokens)


def _analyze_tokens(tokens):
    ending = None
    if tokens and tokens[-1] in STOP_KEYWORDS['both']:
        ending = tokens[-1].value
        del tokens[-1]

    statement = parse_exp(tokens, container=Statement)
    if isinstance(statement, Statement):
        statement.ending = ending
    else:
        statement = Statement([statement], ending=ending)

    return statement


def _analyze_array(tokens, analyze_tokens, tokens_until):
    result = []
    part = []
    first_comma_found = False
    for token in tokens:
        if token == ParserKeyword(','):
            first_comma_found = True
            if not part:
                raise SQFParserError(get_coord(tokens_until), 'Array cannot have an empty element')
            result.append(analyze_tokens(part))
            part = []
        else:
            part.append(token)

    # an empty array is a valid array
    if part == [] and first_comma_found:
        raise SQFParserError(get_coord(tokens_until), 'Array cannot have an empty element')
    elif tokens:
        result.append(analyze_tokens(part))
    return result


def _analyze_define(tokens):
    assert(tokens[0] == Preprocessor('#define'))

    valid_indexes = [i for i in range(len(tokens)) if not isinstance(tokens[i], ParserType)]

    if len(valid_indexes) < 2:
        raise SQFParserError(get_coord(str(tokens[0])), '#define needs at least one argument')
    variable = str(tokens[valid_indexes[1]])
    if len(valid_indexes) == 2:
        return DefineStatement(tokens, variable)
    elif len(valid_indexes) >= 3 and valid_indexes[1] + 1 == valid_indexes[2] and isinstance(tokens[valid_indexes[2]], Statement) and tokens[valid_indexes[2]].parenthesis:
        args = str(tokens[valid_indexes[2]])[1:-1].split(',')
        remaining = tokens[valid_indexes[3]:]
        return DefineStatement(tokens, variable, remaining, args=args)
    elif len(valid_indexes) >= 3:
        remaining = tokens[valid_indexes[2]:]
        return DefineStatement(tokens, variable, remaining)


def find_match_if_def(all_tokens, i, defines, token):
    found = False
    define_statement = None
    arg_indexes = []
    if i + 1 < len(all_tokens) and str(all_tokens[i + 1]) == '(':
        possible_args = defines[str(token)]
        arg_indexes = []
        for arg_number in possible_args:
            if arg_number == 0:
                continue

            for arg_i in range(arg_number + 1):
                if arg_i == arg_number:
                    index = i + 2 + 2 * arg_i - 1
                else:
                    index = i + 2 + 2 * arg_i

                if index >= len(all_tokens):
                    break
                arg_str = str(all_tokens[index])

                if arg_i == arg_number and arg_str != ')':
                    break
                elif not re.match('(.*?)', arg_str):
                    break
                if arg_i != arg_number:
                    arg_indexes.append(index)
            else:
                define_statement = defines[str(token)][arg_number]
                found = True
                break
    elif 0 in defines[str(token)]:
        define_statement = defines[str(token)][0]
        arg_indexes = []
        found = True

    return found, define_statement, arg_indexes


def get_ifdef_variable(tokens, ifdef_i, coord_until_here):
    variable = None
    eol_i = None
    for i, token in enumerate(tokens[ifdef_i:]):
        if type(token) == EndOfLine:
            eol_i = ifdef_i + i
            break
        if type(token) in (Variable, Keyword):
            variable = str(token)
    if variable is not None and eol_i is not None:
        return variable, eol_i
    raise SQFParserError(add_coords(coord_until_here, tokens[:ifdef_i]), '#ifdef statement must contain a variable')


def parse_ifdef_block(expression, defines, coord_until_here):
    """
    Given a IfDefStatement and the defines, converts the statement.tokens into
    a list of tokens that can be analyzed after processing the #ifdef statement.
    `position_until_here` used to compute position of errors
    """
    assert(isinstance(expression, IfDefStatement))
    tokens = expression.tokens
    try:
        ifdef_i = rindex(tokens, Preprocessor('#ifdef'))
        is_ifdef = True
    except ValueError:
        ifdef_i = rindex(tokens, Preprocessor('#ifndef'))
        is_ifdef = False
    try:
        else_i = rindex(tokens, Preprocessor('#else'))
    except ValueError:
        else_i = None
    endif_i = rindex(tokens, Preprocessor('#endif'))
    try:
        # if there is an if_def statement before #endif, the remaining tokens are transferred to it
        nested_if_def = next(i for i, x in enumerate(tokens) if type(x) == IfDefStatement and i < endif_i)
    except StopIteration:
        nested_if_def = None

    variable, eol_i = get_ifdef_variable(tokens, ifdef_i, coord_until_here)

    is_def = (variable in defines)

    replacing_expression = []
    if is_def and is_ifdef or not is_def and not is_ifdef:
        if else_i is None:
            to = endif_i
            if nested_if_def is not None and nested_if_def < endif_i:
                to = nested_if_def + 1
            replacing_expression = tokens[eol_i:to]
        else:
            to = else_i
            if nested_if_def is not None and nested_if_def < else_i:
                to = nested_if_def + 1
            replacing_expression = tokens[eol_i:to]
    elif else_i is not None:
        replacing_expression = tokens[else_i + 1:endif_i]

    try:
        # if there is an if_def statement after the #endif, the remaining tokens are transferred to it
        if_def_i = next(i for i, x in enumerate(tokens) if type(x) == IfDefStatement and i > endif_i)
        next_if_def = tokens[if_def_i]
        replacing_expression.append(next_if_def)
        remaining_tokens = tokens[if_def_i + 1:]
        next_if_def.tokens.extend(remaining_tokens)
        expression._tokens = tokens[:if_def_i + 1]
    except StopIteration:
        replacing_expression += tokens[endif_i + 1:]

    return replacing_expression


def is_finish_ifdef_condition(tokens, lvls):
    return lvls['ifdef'] == sum(1 for token in tokens if token == Preprocessor('#endif')) > 0 and \
        lvls['ifdef_open_close'] == 0


def is_finish_ifdef_parenthesis(token, lvls):
    for lvl_type in ('()', '[]', '{}'):
        if lvls[lvl_type] != 0 and token == ParserKeyword(lvl_type[1]):
            return True
    return False


def finish_ifdef(tokens, all_tokens, start, statements):
    tokens.insert(0, all_tokens[start - 1])  # pick the token that triggered the statement
    assert (len(statements) == 0)
    return IfDefStatement(tokens)


def is_end_statement(token, stop_statement):
    return token in STOP_KEYWORDS[stop_statement] or isinstance(token, EndOfFile)


def parse_block(all_tokens, analyze_tokens, start=0, initial_lvls=None, stop_statement='both', defines=None):
    if not initial_lvls:
        initial_lvls = _LEVELS
    if defines is None:
        defines = defaultdict(dict)
    lvls = initial_lvls.copy()

    statements = []
    tokens = []
    i = start
    if not all_tokens:
        return Statement([]), 0

    while i < len(all_tokens):
        token = all_tokens[i]

        # begin #ifdef controls
        if lvls['ifdef'] and token in OPEN_PARENTHESIS:
            lvls['ifdef_open_close'] += 1

        stop = False
        if token in (Preprocessor('#ifdef'), Preprocessor('#ifndef')):
            stop = True
            lvls['ifdef'] += 1
            expression, size = parse_block(all_tokens, _analyze_simple, i + 1, lvls, stop_statement,
                                           defines=defines)
            lvls['ifdef'] -= 1
            if lvls['ifdef'] == 0:
                assert (isinstance(expression, IfDefStatement))
                replacing_expression = parse_ifdef_block(expression, defines, get_coord(all_tokens[:i - 1]))

                new_all_tokens = sqf.base_type.get_all_tokens(tokens + replacing_expression)

                result, _ = parse_block(new_all_tokens, analyze_tokens, 0, None, stop_statement,
                                        defines=defines)

                expression.prepend(tokens)

                expression = IfDefResult(expression, result.tokens)
                statements.append(expression)

                len_expression = len(expression.get_all_tokens())

                i += len_expression - len(tokens) - 1
                tokens = []
            else:
                tokens.append(expression)
                i += size + 1
        # finish ifdef
        elif is_finish_ifdef_condition(tokens, lvls) and (
                    is_end_statement(token, stop_statement) or
                    is_finish_ifdef_parenthesis(token, lvls)
                ) or lvls['ifdef'] > 1 and token == Preprocessor('#endif'):

            if token != EndOfFile() and token not in CLOSE_PARENTHESIS:
                tokens.append(token)

            if_def = finish_ifdef(tokens, all_tokens, start, statements)
            return if_def, i - start
        # parse during ifdef
        elif lvls['ifdef'] != 0:
            stop = True
            tokens.append(token)

        # end ifdef controls
        if lvls['ifdef'] and token in (STOP_KEYWORDS['single'] + CLOSE_PARENTHESIS):
            lvls['ifdef_open_close'] -= 1
            if lvls['ifdef_open_close'] < 0:
                lvls['ifdef_open_close'] = 0

        if stop:
            pass
        # try to match a #defined and get the arguments
        elif str(token) in defines:  # is a define
            stop, define_statement, arg_indexes = find_match_if_def(all_tokens, i, defines, token)

            if stop:
                arg_number = len(define_statement.args)

                extra_tokens_to_move = 1 + 2 * (arg_number != 0) + 2 * arg_number - 1 * (arg_number != 0)

                replaced_expression = all_tokens[i:i + extra_tokens_to_move]

                # the `all_tokens` after replacement
                replacing_expression = replace_in_expression(define_statement.expression, define_statement.args,
                                                             arg_indexes, all_tokens)

                new_all_tokens = all_tokens[:i - len(tokens)] + tokens + replacing_expression + all_tokens[
                                                                                                i + extra_tokens_to_move:]

                new_start = i - len(tokens)

                expression, size = parse_block(new_all_tokens, analyze_tokens, new_start, lvls,
                                               stop_statement, defines=defines)

                # the all_tokens of the statement before replacement
                original_tokens_taken = len(replaced_expression) - len(replacing_expression) + size

                original_tokens = all_tokens[i - len(tokens):i - len(tokens) + original_tokens_taken]

                if isinstance(expression, Statement):
                    expression = expression.content[0]

                if type(original_tokens[-1]) in (EndOfLine, Comment, EndOfFile):
                    del original_tokens[-1]
                    original_tokens_taken -= 1

                expression = DefineResult(original_tokens, define_statement, expression)
                statements.append(expression)

                i += original_tokens_taken - len(tokens) - 1

                tokens = []
        if stop:
            pass
        elif token == ParserKeyword('['):
            lvls['[]'] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, i + 1, lvls, stop_statement='single', defines=defines)
            lvls['[]'] -= 1
            tokens.append(expression)
            i += size + 1
        elif token == ParserKeyword('('):
            lvls['()'] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, i + 1, lvls, stop_statement, defines=defines)
            lvls['()'] -= 1
            tokens.append(expression)
            i += size + 1
        elif token == ParserKeyword('{'):
            lvls['{}'] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, i + 1, lvls, stop_statement, defines=defines)
            lvls['{}'] -= 1
            tokens.append(expression)
            i += size + 1

        elif token == ParserKeyword(']'):
            if lvls['[]'] == 0:
                raise SQFParenthesisError(get_coord(all_tokens[:i]), 'Trying to close right parenthesis without them opened.')

            if statements:
                if isinstance(statements[0], DefineResult):
                    statements[0]._tokens = [Array(_analyze_array(statements[0]._tokens, analyze_tokens, all_tokens[:i]))]
                    return statements[0], i - start
                else:
                    raise SQFParserError(get_coord(all_tokens[:i]), 'A statement %s cannot be in an array' % Statement(statements))

            return Array(_analyze_array(tokens, analyze_tokens, all_tokens[:i])), i - start
        elif token == ParserKeyword(')'):
            if lvls['()'] == 0:
                raise SQFParenthesisError(get_coord(all_tokens[:i]), 'Trying to close parenthesis without opened parenthesis.')

            if tokens:
                statements.append(analyze_tokens(tokens))

            return Statement(statements, parenthesis=True), i - start
        elif token == ParserKeyword('}'):
            if lvls['{}'] == 0:
                raise SQFParenthesisError(get_coord(all_tokens[:i]), 'Trying to close brackets without opened brackets.')

            if tokens:
                statements.append(analyze_tokens(tokens))

            return Code(statements), i - start
        # end of statement when not in preprocessor states
        elif all(lvls[lvl_type] == 0 for lvl_type in ('#define', '#include')) and is_end_statement(token, stop_statement):
            if type(token) != EndOfFile:
                tokens.append(token)
            if tokens:
                statements.append(analyze_tokens(tokens))

            tokens = []
        elif token in (Preprocessor('#define'), Preprocessor('#include')):
            # notice that `token` is ignored here. It will be picked up in the end
            if tokens:
                # a pre-processor starts a new statement
                statements.append(analyze_tokens(tokens))
                tokens = []

            lvls[token.value] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, i + 1, lvls, stop_statement, defines=defines)
            lvls[token.value] -= 1

            statements.append(expression)
            i += size
        elif type(token) in (EndOfLine, Comment, EndOfFile) and any(lvls[x] != 0 for x in {'#define', '#include'}):
            tokens.insert(0, all_tokens[start - 1])  # pick the token that triggered the statement
            if tokens[0] == Preprocessor('#define'):
                define_statement = _analyze_define(tokens)
                defines[define_statement.variable_name][len(define_statement.args)] = define_statement
                statements.append(define_statement)
            else:
                statements.append(analyze_tokens(tokens))

            return Statement(statements), i - start
        elif type(token) != EndOfFile:
            tokens.append(token)
        i += 1

    if is_finish_ifdef_condition(tokens, lvls):
        return finish_ifdef(tokens, all_tokens, start, statements), i - start

    for lvl_type in ('[]', '()', '{}', 'ifdef'):
        if lvls[lvl_type] != 0:
            message = 'Parenthesis "%s" not closed' % lvl_type[0]
            if lvl_type == 'ifdef':
                message = '#ifdef statement not closed'

            raise SQFParenthesisError(get_coord(all_tokens[:start - 1]), message)

    if tokens:
        statements.append(analyze_tokens(tokens))

    return Statement(statements), i - start


def parse(script):
    tokens = [identify_token(x) for x in parse_strings_and_comments(tokenize(script))]

    result = parse_block(tokens + [EndOfFile()], _analyze_tokens)[0]

    result.set_position((1, 1))

    return result



import re


def tokenize(statement):
    # the len=2 tokens have to be first!
    regex = r'(\\\n|\r\n|>>|\/\*|\*\/|\|\||//|!=|<=|>=|==|\n|\t|[\"\' =:\{\}\(\)\[\];/,\!\/\*\%\^\-\+<>])'
    return list(filter(None, re.split(regex, statement)))



from setuptools import setup

setup(
    name='sqflint',
    version='0.3.2',
    author='Lord Golias',
    author_email='lord.golias1@gmail.com',
    description='Parser, static analyzer and interpreter of SQF (Arma)',
    url='https://github.com/LordGolias/sqf',
    license='BSD',
    packages=['sqf'],
    include_package_data=True,
    scripts=[
        'sqflint.py'
    ],
    entry_points={
        'console_scripts': [
            'sqflint = sqflint:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
    ],
)



import collections

from sqf.types import Keyword


def get_keyword(types_or_values):
    try:
        return next(x.unique_token for x in types_or_values if isinstance(x, Keyword))
    except StopIteration:
        return None


def get_key(types_or_values):
    keyword = get_keyword(types_or_values)
    return keyword, len(types_or_values)


def build_database(expressions):
    database = collections.defaultdict(set)
    for i, exp in enumerate(expressions):
        key = get_key(exp.types_or_values)
        database[key].add(i)

    return database


def values_to_expressions(values, database, expressions):
    key = get_key(values)
    return [expressions[i] for i in database[key]]



from sqf.interpreter import Interpreter, interpret
from sqf.types import Code, Array, String


class Client:
    def __init__(self, simulation):
        self._simulation = simulation
        self._interpreter = Interpreter()
        self._interpreter.client = self
        self._listening_variables = {}  # var_name: code

    @property
    def simulation(self):
        return self._simulation

    def add_listening(self, var_name, code):
        assert(isinstance(var_name, str) and isinstance(code, Code))
        self._listening_variables[var_name] = code

    def execute(self, code):
        interpret(code, self._interpreter)

    def set_variable(self, var_name, value, broadcast=True):
        self._interpreter.set_global_variable(var_name, value)

        if broadcast:
            if var_name in self._listening_variables:
                self._interpreter.execute_code(self._listening_variables[var_name],
                                               extra_scope={'_this': Array([String('"'+var_name+'"'), value])})

    @property
    def is_server(self):
        return self._simulation.server == self

    @property
    def is_dedicated(self):
        return self._simulation.is_dedicated


class Simulation:

    def __init__(self, is_dedicated=True):
        self._is_dedicated = is_dedicated
        self.server = Client(self)
        self._clients = []

        self._broadcasted = {}

    @property
    def is_dedicated(self):
        return self._is_dedicated

    @property
    def clients(self):
        return self._clients

    def add_client(self, client):
        assert self._is_dedicated

        self._clients.append(client)

        for var_name in self._broadcasted:
            client.set_variable(var_name, self._broadcasted[var_name], broadcast=False)

        return len(self._clients) - 1

    def broadcast(self, var_name, value, client_id=None):
        # client_id=None => to all
        # client_id=-1 => to the server
        if client_id is None:
            self._broadcasted[var_name] = value
            for client in self._clients + [self.server]:
                client.set_variable(var_name, value)
        elif client_id == -1:
            self.server.set_variable(var_name, value)
        else:
            self._clients[client_id].set_variable(var_name, value)
